'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.calculatePopoverPosition = undefined;

var _assign2 = require('lodash/assign');

var _assign3 = _interopRequireDefault(_assign2);

var _prefix = require('prefix');

var _prefix2 = _interopRequireDefault(_prefix);

var _getWindow = require('./get-window');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

// Not currently designed to deal with
// - iframes
// - more than one level of sub-document scrolling
// - horizontal scrolling

var PLACEMENT_TOP = 'top';
var PLACEMENT_BOTTOM = 'bottom';
var PLACEMENT_LEFT = 'left';
var PLACEMENT_RIGHT = 'right';
var ALIGNMENT_TOP = 'top';
var ALIGNMENT_BOTTOM = 'bottom';
var ALIGNMENT_LEFT = 'left';
var ALIGNMENT_RIGHT = 'right';
var ALIGNMENT_CENTER = 'center';
var POINTER_DIRECTION_DOWNWARD = 'downward';
var POINTER_DIRECTION_UPWARD = 'upward';
var POINTER_DIRECTION_LEFTWARD = 'leftward';
var POINTER_DIRECTION_RIGHTWARD = 'rightward';

var mapPlacementToPointerDirection = {};
mapPlacementToPointerDirection[PLACEMENT_TOP] = POINTER_DIRECTION_DOWNWARD;
mapPlacementToPointerDirection[PLACEMENT_BOTTOM] = POINTER_DIRECTION_UPWARD;
mapPlacementToPointerDirection[PLACEMENT_LEFT] = POINTER_DIRECTION_RIGHTWARD;
mapPlacementToPointerDirection[PLACEMENT_RIGHT] = POINTER_DIRECTION_LEFTWARD;

// Default, without making coercion mistakes
function def(value, def) {
  if (value !== undefined) return value;
  return def;
}

/**
 * @typedef {Object} PopoverPositionData
 * @property {string} placement
 * @property {string} alignment
 * @property {number} topOffset
 * @property {number} leftOffset
 * @property {Object} pointerPositioningStyle
 * @property {Object} bodyPositioningStyle
 */

/**
 * Calculate positioning for a popover. Returns PopoverPositionData.
 *
 * @param {Object} options
 * @param {'top' | 'bottom' | 'left' | 'right'} options.placement
 * @param {'top' | 'bottom' | 'left' | 'right' | 'center' } options.alignment
 * @param {Function} options.getAnchorElement - Returns anchor element.
 * @param {Function} options.getPopoverElement - Returns popover element.
 * @param {Function} [options.getContainingElement] - Returns containing element.
 * @param {number} [options.offsetFromAnchor=5]
 * @param {number} [options.offsetFromEdge]
 * @param {boolean} [options.containWithinViewport=false]
 * @param {Function} [options.getScrollableParentElement] - Returns scrollable parent element.
 * @param {boolean} [options.allowPlacementAxisChange=true]
 * @param {boolean} [options.hasPointer=true]
 * @param {number} [options.pointerBase=6]
 * @param {number} [options.pointerAltitude=12]
 * @param {string} [options.pointerColor='#fff']
 * @param {boolean} [options.hideWhenAnchorIsOffscreen=false]
 * @return {PopoverPositionData}
 */
function calculatePopoverPosition(options) {
  var _bodyPositioningStyle, _pointerPositioningSt, _pointerPositioningSt2, _pointerPositioningSt3, _pointerPositioningSt4;

  var win = (0, _getWindow.getWindow)();
  var docEl = win.document.documentElement;
  var scrollY = win.pageYOffset;

  var placement = options.placement,
      alignment = options.alignment;

  var anchor = options.getAnchorElement();
  if (!anchor) throw new Error('No anchor element provided');
  var containingElementRect = options.getContainingElement ? options.getContainingElement().getBoundingClientRect() : null;
  var offsetFromEdgeDefault = containingElementRect ? 0 : 5;
  var offsetFromEdge = def(options.offsetFromEdgem, offsetFromEdgeDefault);
  var offsetFromAnchor = def(options.offsetFromAnchor, 5);
  var containWithinViewport = def(options.containWithinViewport, false);
  var hasPointer = def(options.hasPointer, true);
  var pointerAltitude = hasPointer === false ? 0 : def(options.pointerAltitude, 6);
  var pointerBase = hasPointer === false ? 0 : def(options.pointerBase, 12);
  var bodyOffsetFromAnchor = offsetFromAnchor + pointerAltitude;
  var pointerColor = def(options.pointerColor, '#fff');
  var allowPlacementAxisChange = def(options.allowPlacementAxisChange, true);
  var availableWidth = docEl.scrollWidth;

  // If the document element's top position has been manipulated, we'll need to adjust accordingly
  var documentTopOffset = 0;
  var documentPosition = docEl.style.position;
  if (documentPosition === 'fixed' || documentPosition === 'absolute') {
    documentTopOffset = parseFloat(docEl.style.top);
  }

  var availableHeight = void 0;
  if (options.containWithinViewport) {
    availableHeight = win.innerHeight;
  } else {
    availableHeight = docEl.scrollHeight;
  }

  var topBoundary = void 0;
  if (containingElementRect) {
    topBoundary = containingElementRect.top + scrollY + offsetFromEdge;
  } else if (containWithinViewport === true) {
    topBoundary = scrollY + offsetFromEdge;
  } else {
    topBoundary = offsetFromEdge;
  }

  var bottomBoundary = void 0;
  if (containingElementRect) {
    bottomBoundary = containingElementRect.bottom + scrollY - offsetFromEdge;
  } else if (containWithinViewport === true) {
    bottomBoundary = scrollY + availableHeight - offsetFromEdge;
  } else {
    bottomBoundary = availableHeight - offsetFromEdge;
  }

  var anchorRect = anchor.getBoundingClientRect();
  var bodyElement = options.getPopoverElement();
  if (!bodyElement) throw new Error('No body element provided');
  var bodyElementRect = bodyElement.getBoundingClientRect();
  var spaceLeftOfAnchor = anchorRect.left - offsetFromEdge;
  var spaceRightOfAnchor = availableWidth - anchorRect.right - offsetFromEdge;
  var spaceAboveAnchor = anchorRect.top - offsetFromEdge;
  var spaceBelowAnchor = availableHeight - anchorRect.bottom - offsetFromEdge;
  var anchorAbsoluteTop = anchorRect.top + scrollY;
  var anchorAbsoluteBottom = anchorRect.bottom + scrollY;
  var sufficientSpaceIfPlacedOnLeft = bodyElementRect.width + bodyOffsetFromAnchor <= spaceLeftOfAnchor;
  var sufficientSpaceIfPlacedOnRight = bodyElementRect.width + bodyOffsetFromAnchor <= spaceRightOfAnchor;
  var sufficientSpaceIfPlacedAbove = bodyElementRect.height + bodyOffsetFromAnchor <= spaceAboveAnchor;
  var sufficientSpaceIfPlacedBelow = bodyElementRect.height + bodyOffsetFromAnchor <= spaceBelowAnchor;
  var maxLeftOffset = Math.max(availableWidth - offsetFromEdge - bodyElementRect.width, offsetFromEdge);
  var maxWidth = availableWidth - 2 * offsetFromEdge;
  var maxHeight = availableHeight - 2 * offsetFromEdge;

  var calculatedPlacement = placement;
  var calculatedAlignment = alignment;

  if (placement === PLACEMENT_TOP || placement === PLACEMENT_BOTTOM) {
    if (placement === PLACEMENT_BOTTOM && !sufficientSpaceIfPlacedBelow) {
      if (sufficientSpaceIfPlacedAbove) {
        calculatedPlacement = PLACEMENT_TOP;
      } else if (allowPlacementAxisChange && sufficientSpaceIfPlacedOnLeft) {
        calculatedPlacement = PLACEMENT_LEFT;
        calculatedAlignment = ALIGNMENT_CENTER;
      } else if (allowPlacementAxisChange && sufficientSpaceIfPlacedOnRight) {
        calculatedPlacement = PLACEMENT_RIGHT;
        calculatedAlignment = ALIGNMENT_CENTER;
      } else if (spaceAboveAnchor > spaceBelowAnchor) {
        calculatedPlacement = PLACEMENT_TOP;
      }
    }

    if (placement === PLACEMENT_TOP && !sufficientSpaceIfPlacedAbove) {
      if (sufficientSpaceIfPlacedBelow) {
        calculatedPlacement = PLACEMENT_BOTTOM;
      } else if (allowPlacementAxisChange && sufficientSpaceIfPlacedOnLeft) {
        calculatedPlacement = PLACEMENT_LEFT;
        calculatedAlignment = ALIGNMENT_CENTER;
      } else if (allowPlacementAxisChange && sufficientSpaceIfPlacedOnRight) {
        calculatedPlacement = PLACEMENT_RIGHT;
        calculatedAlignment = ALIGNMENT_CENTER;
      } else if (spaceBelowAnchor > spaceAboveAnchor) {
        calculatedPlacement = PLACEMENT_BOTTOM;
      }
    }
  }

  if (placement === PLACEMENT_LEFT || placement === PLACEMENT_RIGHT) {
    if (placement === PLACEMENT_LEFT && !sufficientSpaceIfPlacedOnLeft) {
      if (sufficientSpaceIfPlacedOnRight) {
        calculatedPlacement = PLACEMENT_RIGHT;
      } else if (allowPlacementAxisChange && !sufficientSpaceIfPlacedBelow && sufficientSpaceIfPlacedAbove) {
        calculatedPlacement = PLACEMENT_TOP;
        calculatedAlignment = ALIGNMENT_CENTER;
      } else if (allowPlacementAxisChange) {
        calculatedPlacement = PLACEMENT_BOTTOM;
        calculatedAlignment = ALIGNMENT_CENTER;
      } else if (spaceRightOfAnchor > spaceLeftOfAnchor) {
        calculatedPlacement = PLACEMENT_RIGHT;
      }
    }

    if (placement === PLACEMENT_RIGHT && !sufficientSpaceIfPlacedOnRight) {
      if (sufficientSpaceIfPlacedOnLeft) {
        calculatedPlacement = PLACEMENT_LEFT;
      } else if (allowPlacementAxisChange && !sufficientSpaceIfPlacedBelow && sufficientSpaceIfPlacedAbove) {
        calculatedPlacement = PLACEMENT_TOP;
        calculatedAlignment = ALIGNMENT_CENTER;
      } else if (allowPlacementAxisChange) {
        calculatedPlacement = PLACEMENT_BOTTOM;
        calculatedAlignment = ALIGNMENT_CENTER;
      } else if (spaceLeftOfAnchor > spaceRightOfAnchor) {
        calculatedPlacement = PLACEMENT_LEFT;
      }
    }
  }

  // Having determined the placement and alignment,
  // translate those into actual offsetFromAnchors
  // =========================================================

  var topOffset = void 0;
  var leftOffset = void 0;

  if (calculatedPlacement === PLACEMENT_TOP) {
    topOffset = anchorAbsoluteTop - bodyElementRect.height - bodyOffsetFromAnchor;
  } else if (calculatedPlacement === PLACEMENT_BOTTOM) {
    topOffset = anchorAbsoluteBottom + bodyOffsetFromAnchor;
  } else if (calculatedAlignment === ALIGNMENT_TOP) {
    topOffset = anchorAbsoluteTop;
  } else if (calculatedAlignment === ALIGNMENT_BOTTOM) {
    topOffset = anchorAbsoluteBottom - bodyElementRect.height;
  } else {
    // Only situation left is left/right placed with alignment center
    topOffset = anchorAbsoluteTop + anchorRect.height / 2 - bodyElementRect.height / 2;
  }
  topOffset = Math.max(offsetFromEdge, Math.round(topOffset));
  // topOffset = Math.min(maxTopOffset, topOffset);
  if (topOffset < topBoundary) topOffset = topBoundary;
  if (topOffset + bodyElementRect.height > bottomBoundary) topOffset = bottomBoundary - bodyElementRect.height;
  topOffset -= documentTopOffset;

  if (calculatedPlacement === PLACEMENT_LEFT) {
    leftOffset = anchorRect.left - bodyElementRect.width - bodyOffsetFromAnchor;
  } else if (calculatedPlacement === PLACEMENT_RIGHT) {
    leftOffset = anchorRect.right + bodyOffsetFromAnchor;
  } else if (calculatedAlignment === ALIGNMENT_LEFT) {
    leftOffset = anchorRect.left;
  } else if (calculatedAlignment === ALIGNMENT_RIGHT) {
    leftOffset = anchorRect.right - bodyElementRect.width;
  } else {
    // Only situation left is top/bottom placed with alignment center
    leftOffset = anchorRect.left + anchorRect.width / 2 - bodyElementRect.width / 2;
  }
  leftOffset = Math.max(offsetFromEdge, Math.round(leftOffset));
  leftOffset = Math.min(maxLeftOffset, leftOffset);

  var prefixedTransform = _prefix2.default.dash('transform');

  var bodyPositioningStyle = (_bodyPositioningStyle = {}, _defineProperty(_bodyPositioningStyle, prefixedTransform, 'translate3d(' + leftOffset + 'px, ' + topOffset + 'px, 0)'), _defineProperty(_bodyPositioningStyle, 'max-width', maxWidth + 'px'), _defineProperty(_bodyPositioningStyle, 'max-height', maxHeight + 'px'), _bodyPositioningStyle);

  // Position and style the pointer
  // =========================================================

  var anchorHorizontalCenter = anchorRect.left + anchorRect.width / 2;
  var anchorVerticalCenter = anchorAbsoluteTop + anchorRect.height / 2;
  var verticalPointerLeftOffset = Math.round(anchorHorizontalCenter - pointerBase / 2);
  var horizontalPointerTopOffset = Math.round(anchorVerticalCenter - pointerBase / 2);
  var helperBorder = pointerBase / 2 + 'px solid transparent';
  var coloredBorder = pointerAltitude + 'px solid ' + pointerColor;
  var pointerDirection = mapPlacementToPointerDirection[calculatedPlacement];

  var pointerPositioningStyle = {};
  if (hasPointer === false) {
    pointerPositioningStyle = { display: 'none' };
  } else {
    switch (pointerDirection) {
      case POINTER_DIRECTION_UPWARD:
        pointerPositioningStyle = (_pointerPositioningSt = {}, _defineProperty(_pointerPositioningSt, prefixedTransform, 'translate3d(' + verticalPointerLeftOffset + 'px, ' + Math.round(anchorAbsoluteTop + anchorRect.height + offsetFromAnchor - documentTopOffset) + 'px, 0)'), _defineProperty(_pointerPositioningSt, 'border-top', 0), _defineProperty(_pointerPositioningSt, 'border-bottom', coloredBorder), _defineProperty(_pointerPositioningSt, 'border-left', helperBorder), _defineProperty(_pointerPositioningSt, 'border-right', helperBorder), _pointerPositioningSt);
        break;
      case POINTER_DIRECTION_DOWNWARD:
        pointerPositioningStyle = (_pointerPositioningSt2 = {}, _defineProperty(_pointerPositioningSt2, prefixedTransform, 'translate3d(' + verticalPointerLeftOffset + 'px, ' + Math.round(anchorAbsoluteTop - offsetFromAnchor - pointerAltitude - documentTopOffset) + 'px, 0)'), _defineProperty(_pointerPositioningSt2, 'border-top', coloredBorder), _defineProperty(_pointerPositioningSt2, 'border-bottom', 0), _defineProperty(_pointerPositioningSt2, 'border-left', helperBorder), _defineProperty(_pointerPositioningSt2, 'border-right', helperBorder), _pointerPositioningSt2);
        break;
      case POINTER_DIRECTION_LEFTWARD:
        pointerPositioningStyle = (_pointerPositioningSt3 = {}, _defineProperty(_pointerPositioningSt3, prefixedTransform, 'translate3d(' + Math.round(anchorRect.right + offsetFromAnchor) + 'px, ' + (horizontalPointerTopOffset - documentTopOffset) + 'px, 0)'), _defineProperty(_pointerPositioningSt3, 'border-top', helperBorder), _defineProperty(_pointerPositioningSt3, 'border-bottom', helperBorder), _defineProperty(_pointerPositioningSt3, 'border-left', 0), _defineProperty(_pointerPositioningSt3, 'border-right', coloredBorder), _pointerPositioningSt3);
        break;
      case POINTER_DIRECTION_RIGHTWARD:
        pointerPositioningStyle = (_pointerPositioningSt4 = {}, _defineProperty(_pointerPositioningSt4, prefixedTransform, 'translate3d(' + Math.round(anchorRect.left - offsetFromAnchor - pointerAltitude) + 'px, ' + (horizontalPointerTopOffset - documentTopOffset) + 'px, 0)'), _defineProperty(_pointerPositioningSt4, 'border-top', helperBorder), _defineProperty(_pointerPositioningSt4, 'border-bottom', helperBorder), _defineProperty(_pointerPositioningSt4, 'border-left', coloredBorder), _defineProperty(_pointerPositioningSt4, 'border-right', 0), _pointerPositioningSt4);
        break;
    }
  }

  // Determine whether to hide things or not
  // =========================================================

  var hideStyles = {
    visibility: 'hidden',
    zIndex: -1
  };

  var showStyles = {
    visibility: 'visible',
    zIndex: 1
  };

  function hideThings() {
    (0, _assign3.default)(pointerPositioningStyle, hideStyles);
    if (options.hideWhenAnchorIsOffscreen) {
      (0, _assign3.default)(bodyPositioningStyle, hideStyles);
    }
  }

  function showThings() {
    (0, _assign3.default)(pointerPositioningStyle, showStyles);
    if (options.hideWhenAnchorIsOffscreen) {
      (0, _assign3.default)(bodyPositioningStyle, showStyles);
    }
  }

  var scrollableParent = options.getScrollableParentElement !== undefined ? options.getScrollableParentElement() : null;

  var scrollableParentTop = void 0;
  var scrollableParentBottom = void 0;
  if (scrollableParent) {
    if (scrollableParent === win) {
      scrollableParentTop = options.containWithinViewport ? scrollY : 0;
      scrollableParentBottom = options.containWithinViewport ? scrollY + win.innerHeight : availableHeight;
    } else {
      var scrollableParentRect = scrollableParent.getBoundingClientRect();
      scrollableParentTop = scrollableParentRect.top + scrollY;
      scrollableParentBottom = scrollableParentRect.bottom + scrollY;
    }

    var offsetFromScrollableParentTopEdge = Math.round(anchorVerticalCenter - pointerBase / 2) - 2;
    var offsetFromScrollableParentBottomEdge = Math.round(anchorVerticalCenter + pointerBase / 2) + 2;
    var onLeftOrRide = calculatedPlacement === PLACEMENT_LEFT || calculatedPlacement === PLACEMENT_RIGHT;

    if (calculatedPlacement === PLACEMENT_TOP && scrollableParentTop > anchorAbsoluteTop || calculatedPlacement === PLACEMENT_BOTTOM && scrollableParentBottom < anchorAbsoluteTop || !onLeftOrRide && scrollableParentBottom < anchorAbsoluteTop || !onLeftOrRide && scrollableParentTop > anchorAbsoluteBottom || onLeftOrRide && (
    // Offset the allowed top/bottom to prevent the pointer triangle from appearing
    // beyond the border curves
    scrollableParentTop + pointerBase / 2 + 1 > offsetFromScrollableParentTopEdge || scrollableParentBottom - pointerBase / 2 - 1 < offsetFromScrollableParentBottomEdge)) {
      hideThings();
    } else {
      showThings();
    }
  }

  // Return something of value
  // =========================================================

  return {
    placement: calculatedPlacement,
    alignment: calculatedAlignment,
    topOffset: topOffset,
    leftOffset: leftOffset,
    pointerPositioningStyle: pointerPositioningStyle,
    bodyPositioningStyle: bodyPositioningStyle
  };
}

exports.calculatePopoverPosition = calculatePopoverPosition;